"""
panel
~~~~~

Base classes for objects that display data in an area in a terminal.
"""
from collections.abc import Callable
from typing import Optional

from blessed import Terminal
from blessed.keyboard import Keystroke

from thurible.util import Box, get_terminal


# Exceptions.
class InvalidDimensionsError(Exception):
    """The parameters that determine the relative width or height
    must add up to one.
    """


class InvalidTitleAlignmentError(Exception):
    """The value given for the title alignment was invalid."""


class NoFrameTypeForFrameError(Exception):
    """A paramters that requires a frame type was set without setting a
    frame type.
    """


class PanelPaddingAndAlignmentSetError(Exception):
    """You cannot set both panel padding and alignment."""


# Base classes.
class Message:
    """A base class to allow all messages to be identified."""


class Panel:
    """Create a new :class:`Panel` object. This class serves as a parent
    class for all panels, providing the core code relating to the
    area the panel fills in the terminal window.

    :param height: (Optional.) The height of the pane.
    :param width: (Optional.) The width of the pane.
    :param term: (Optional.) A :class:`blessed.Terminal` instance for
        formatting text for terminal display.
    :param origin_y: (Optional.) The terminal row for the top of the
        panel.
    :param origin_x: (Optional.) The terminal column for the left
        side of the panel.
    :param fg: (Optional.) A string describing the foreground color
        of the pane. See the documentation for :mod:`blessed` for more
        detail on the available options.
    :param bg: (Optional.) A string describing the background color
        of the pane. See the documentation for :mod:`blessed` for more
        detail on the available options.
    :param panel_pad_top: (Optional.) Distance between the
        Y origin of the panel and the top of the interior of the
        panel. It is a percentage expressed as a :class:`float`
        between 0.0 and 1.0, inclusive. See :ref:`sizing` for more
        information.
    :param panel_relative_height: (Optional.) The height of the
        interior of the panel in comparison of the full `height` of
        the panel. It is a percentage expressed as a :class:`float`
        between 0.0 and 1.0, inclusive. See :ref:`sizing` for more
        information.
    :param panel_pad_bottom: (Optional.) Distance between the
        full height of the panel and the interior of the panel.
        It is a percentage expressed as a :class:`float` between 0.0
        and 1.0, inclusive. See :ref:`sizing` for more information.
    :param panel_pad_left: (Optional.) Distance between the
        X origin of the panel and the left side of the interior
        of the panel. It is a percentage expressed as a :class:`float`
        between 0.0 and 1.0, inclusive. See :ref:`sizing` for more
        information.
    :param panel_relative_width: (Optional.) The width of the
        interior of the panel in comparison of the full width of
        the panel. It is a percentage expressed as a :class:`float`
        between 0.0 and 1.0, inclusive. See :ref:`sizing` for more
        information.
    :param panel_pad_right: (Optional.) Distance between the full
        width of the panel and the right side of the interior of
        the panel. It is a percentage expressed as a :class:`float`
        between 0.0 and 1.0, inclusive. See :ref:`sizing` for more
        information.
    :param panel_align_h: (Optional.) If the interior of the panel
        is smaller than the full width of the panel, this sets
        how the interior of the panel is aligned within the full
        height. It is a percentage expressed as a :class:`float`
        between 0.0 and 1.0, inclusive. See :ref:`sizing` for more
        information.
    :param panel_align_v: (Optional.) If the interior of the panel
        is smaller than the full width of the panel, this sets
        how the interior of the panel is aligned within the full
        height. It is a percentage expressed as a :class:`float`
        between 0.0 and 1.0, inclusive. See :ref:`sizing` for more
        information.
    :return: None.
    :rtype: NoneType
    """
    # Magic methods.
    def __init__(
        self,
        height: Optional[int] = None,
        width: Optional[int] = None,
        term: Optional[Terminal] = None,
        origin_y: Optional[int] = None,
        origin_x: Optional[int] = None,
        bg: str = '',
        fg: str = '',
        panel_pad_bottom: Optional[float] = None,
        panel_pad_left: Optional[float] = None,
        panel_pad_right: Optional[float] = None,
        panel_pad_top: Optional[float] = None,
        panel_relative_height: Optional[float] = None,
        panel_relative_width: Optional[float] = None,
        panel_align_h: Optional[str] = None,
        panel_align_v: Optional[str] = None
    ) -> None:
        self.term = term if term else get_terminal()
        self._set_relative_horizontal_dimensions(
            panel_pad_left,
            panel_relative_width,
            panel_pad_right,
            panel_align_h
        )
        self._set_relative_vertical_dimensions(
            panel_pad_top,
            panel_relative_height,
            panel_pad_bottom,
            panel_align_v
        )
        self.height = height if height is not None else self.term.height
        self.width = width if width is not None else self.term.width
        self.origin_y = origin_y if origin_y else 0
        self.origin_x = origin_x if origin_x else 0
        self.bg = bg
        self.fg = fg

        # Private attributes.
        self._active_keys: dict[str, Callable] = {}

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.height == other.height
            and self.width == other.width
            and self.origin_y == other.origin_y
            and self.origin_x == other.origin_x
            and self.bg == other.bg
            and self.fg == other.fg
        )

    def __str__(self) -> str:
        """Return a string that will draw the entire display."""
        result = ''
        result += self.clear_contents()
        return result

    # Properties
    @property
    def active_keys(self) -> dict[str, Callable]:
        """The key presses the class will react to and the handler that
        acts on that key press.

        :return: A :class:`dict` object where the keys are the representation
            of the :class:`blessed.keyboard.Keystroke` object emitted when
            the key is pressed and the values are the action handler
            methods called when the key is pressed.
        :rtype: dict
        """
        return self._active_keys.copy()

    @property
    def inner_height(self) -> int:
        """The number of rows in the terminal contained within the
        interior of the panel.

        :return: An :class:`int` object.
        :rtype: int
        """
        height = self.height
        height -= self._panel_pad_offset_top
        height -= self._panel_pad_offset_bottom
        return height

    @property
    def inner_width(self) -> int:
        """The number of columns in the terminal contained within the
        interior of the panel.

        :return: An :class:`int` object.
        :rtype: int
        """
        width = self.width
        width -= self._panel_pad_offset_left
        width -= self._panel_pad_offset_right
        return width

    @property
    def inner_x(self) -> int:
        """The left-most column in the terminal of the interior of the
        panel.

        :return: An :class:`int` object.
        :rtype: int
        """
        return self.origin_x + self._panel_pad_offset_left

    @property
    def inner_y(self) -> int:
        """The top-most row in the terminal of the interior of the panel.

        :return: An :class:`int` object.
        :rtype: int
        """
        return self.origin_y + self._panel_pad_offset_top

    @property
    def _panel_pad_offset_bottom(self) -> int:
        offset = self.height * self.panel_pad_bottom
        return int(offset)

    @property
    def _panel_pad_offset_left(self) -> int:
        offset = self.width * self.panel_pad_left
        return int(offset)

    @property
    def _panel_pad_offset_right(self) -> int:
        offset = self.width * self.panel_pad_right
        return int(offset)

    @property
    def _panel_pad_offset_top(self) -> int:
        offset = self.height * self.panel_pad_top
        return int(offset)

    # Public methods.
    def action(self, key: Keystroke) -> tuple[str, str]:
        """Act on a keystroke typed by the user.

        :param key: A :class:`blessed.keyboard.Keystroke` object
            representing the key pressed by the user.
        :return: A :class:`tuple` object containing two :class:`str`
            objects. The first string is any data that needs to be sent
            to the application. The second string contains any updates
            needed to be made to the terminal display.
        :rtype: tuple
        """
        data = str(key)
        update = ''
        return (data, update)

    def clear_contents(self) -> str:
        """Clear the interior area of the panel.

        :return: A :class:`str` object containing the update needed to
            be made to the terminal display.
        :rtype: str
        """
        # Set up.
        height = self.inner_height
        width = self.inner_width
        y = self.inner_y
        x = self.inner_x
        color = self._get_color(self.fg, self.bg)
        result = color

        # Create the clearing string and return.
        for i in range(height):
            result += self.term.move(y + i, x) + ' ' * width
        if color:
            result += self.term.normal
        return result

    def register_key(self, key: str, handler: Callable) -> None:
        """Declare the key presses the class will react to, and define
        the action the class will take when that key is pressed.

        :param key: The name of the key pressed as returned by the
            representation of the :class:`blessed.keyboard.Keystroke`
            emitted by the key press.
        :param handler: And action handler to invoke when the key is
            pressed. An action handler is a function that takes an
            optional :class:`blessed.keyboard.Keystroke` object and
            returns a string that contains any changes that need to be
            made to the terminal display as a result of the key press.
        :return: None.
        :rtype: NoneType
        """
        items = [getattr(self, name) for name in self.__dict__]
        if handler not in items:
            setattr(self, key, handler)
        self._active_keys[key] = handler

    def update(self, msg: Message) -> str:
        """Act on a message sent by the application.

        :param msg: A message sent by the application.
        :return: A :class:`str` object containing any updates needed to
            be made to the terminal display.
        :rtype: str
        """
        return ''

    # Private helper methods.
    def _get_color(self, fg: str = '', bg: str = '') -> str:
        color = fg
        if color and bg:
            color += f'_on_{bg}'
        elif bg:
            color += f'on_{bg}'
        return getattr(self.term, color)

    def _set_relative_dimenstion(
        self,
        left: Optional[float] = None,
        width: Optional[float] = None,
        right: Optional[float] = None,
        align: Optional[str] = None,
        align_default: str = 'center',
        alignments: tuple[str, str, str] = ('left', 'center', 'right'),
        attr_names: tuple[str, str, str] = (
            'panel_pad_left',
            'panel_relative_width',
            'panel_pad_right',
        )
    ) -> tuple[float, float, float, str]:
        # This function needs to check which parameters are None a lot.
        # So, the answer to that is stored in this tuple to help
        # reduce the verbosity of the rest of the function.
        were_set = (left is not None, width is not None, right is not None)
        LEFT, WIDTH, RIGHT = 0, 1, 2

        # Since we aren't directly checking these values to see if they
        # are None, mypy will get confused. So, replace Nones with 0.0
        # to keep mypy happy.
        left = 0.0 if left is None else left
        width = 0.0 if width is None else width
        right = 0.0 if right is None else right

        # If both padding and alignment are set, raise an exception
        # because the intended behavior of the panel would be ambiguous.
        if align is not None and any((were_set[LEFT], were_set[RIGHT])):
            msg = 'Cannot set both panel padding and panel alignment.'
            raise PanelPaddingAndAlignmentSetError(msg)

        # If only width was set and align wasn't, set align to the default.
        elif align is None:
            align = align_default

        # If none are set, use the default values.
        if not any(were_set):
            left = 0.0
            right = 0.0
            width = 1.0

        # If all three values are set, they must add up to one.
        elif all(were_set) and left + right + width != 1.0:
            msg = (
                f'If {attr_names[LEFT]}, {attr_names[RIGHT]}, and '
                f'{attr_names[WIDTH]} are set, the sum of '
                'the three must equal one. The given values were: '
                f'{attr_names[LEFT]}={left}, '
                f'{attr_names[RIGHT]}={right}, '
                f'and {attr_names[WIDTH]}={width}.'
            )
            raise InvalidDimensionsError(msg)

        # If only left was set, the rest goes to width.
        elif sum(were_set) == 1 and were_set[LEFT]:
            width = 1.0 - left
            right = 0.0

        # If only width was set, the padding depends on the alignment.
        elif sum(were_set) == 1 and were_set[WIDTH]:
            total = 1.0 - width
            if align == alignments[LEFT]:
                left = 0.0
                right = total
            elif align == alignments[RIGHT]:
                left = total
                right = 0.0
            else:
                left = total / 2
                right = total / 2

        # If only right was set, the rest goes to width.
        elif sum(were_set) == 1 and were_set[RIGHT]:
            left = 0.0
            width = 1.0 - right

        # If only left wasn't set, width is everything that remains.
        elif sum(were_set) == 2 and not were_set[LEFT]:
            left = 1.0 - width - right

        # If only width wasn't set, width is everything that's not pad.
        elif sum(were_set) == 2 and not were_set[WIDTH]:
            width = 1.0 - left - right

        # If only right wasn't set, right is everything that remains.
        elif sum(were_set) == 2 and not were_set[RIGHT]:
            right = 1.0 - left - width

        # Return the values.
        return left, width, right, align

    def _set_relative_horizontal_dimensions(
        self,
        left: Optional[float] = None,
        width: Optional[float] = None,
        right: Optional[float] = None,
        align: Optional[str] = None,
    ) -> None:
        """Ensure the horizontal relative dimensions are set correctly."""
        # Calculate the correct values.
        left, width, right, align = self._set_relative_dimenstion(
            left,
            width,
            right,
            align
        )

        # Set the attributes.
        self.panel_pad_left = left
        self.panel_relative_width = width
        self.panel_pad_right = right
        self.panel_align_h = align

    def _set_relative_vertical_dimensions(
        self,
        top: Optional[float] = None,
        height: Optional[float] = None,
        bottom: Optional[float] = None,
        align: Optional[str] = None
    ) -> None:
        # Calculate the correct values.
        top, height, bottom, align = self._set_relative_dimenstion(
            top,
            height,
            bottom,
            align,
            align_default='middle',
            alignments=('top', 'middle', 'bottom'),
            attr_names=(
                'panel_pad_top',
                'panel_relative_height',
                'panel_pad_bottom',
            )
        )

        # Set the attributes.
        self.panel_pad_top = top
        self.panel_relative_height = height
        self.panel_pad_bottom = bottom
        self.panel_align_v = align


# Protocols.
class Frame(Panel):
    """Create a new :class:`thurible.panel.Frame` object. This class
    serves as a parent class for all panels that can have a frame
    surrounding the interior of the panel. As a subclass of
    :class:`thurible.panel.Panel`, it can also take those parameters
    and has those public methods.

    :param frame_type: (Optional.) If a string, the string determines
        the frame used for the pane. If None, the pane doesn't have a
        frame.
    :param frame_fg: (Optional.) A string describing the foreground
        color of the frame. See the documentation for :mod:`blessed`
        for more detail on the available options. If `fg` is set and
        this is not, the frame will have the `fg` color.
    :param frame_bg: (Optional.) A string describing the background
        color of the frame. See the documentation for `blessed` for
        more detail on the available options. If `bg` is set and
        this is not, the frame will have the `bg` color.
    :return: None.
    :rtype: NoneType
    """
    def __init__(
        self,
        frame_type: Optional[str] = None,
        frame_bg: str = '',
        frame_fg: str = '',
        *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.frame_type = frame_type
        self.frame_bg = frame_bg
        self.frame_fg = frame_fg

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            super().__eq__(other)
            and self.frame_type == other.frame_type
            and self.frame_bg == other.frame_bg
            and self.frame_fg == other.frame_fg
        )

    def __str__(self) -> str:
        """Return a string that will draw the entire display."""
        result = super().__str__()
        result += self.frame
        return result

    # Properties
    @property
    def frame(self) -> str:
        """A string that will print panel's frame in a terminal.

        :return: A :class:`str` object.
        :rtype: str
        """
        # Handle frame coloration.
        bg = self.frame_bg if self.frame_bg else self.bg
        fg = self.frame_fg if self.frame_fg else self.fg

        # Create the frame string and return.
        result = ''
        if self.frame_type is not None:
            result += self._frame(
                frame_type=self.frame_type,
                height=self.frame_height,
                width=self.frame_width,
                origin_y=self.frame_origin_y,
                origin_x=self.frame_origin_x,
                foreground=fg,
                background=bg
            )
        return result

    @property
    def frame_height(self) -> int:
        """The height in rows of the frame in the terminal.

        :return: A :class:`str` object.
        :rtype: str
        """
        return super().inner_height

    @property
    def frame_width(self) -> int:
        """The width in columns of the frame in the terminal.

        :return: A :class:`str` object.
        :rtype: str
        """
        return super().inner_width

    @property
    def frame_origin_x(self) -> int:
        """The left-most column of the frame in the terminal.

        :return: A :class:`str` object.
        :rtype: str
        """
        return super().inner_x

    @property
    def frame_origin_y(self) -> int:
        """The top-most row of the frame in the terminal.

        :return: A :class:`str` object.
        :rtype: str
        """
        return super().inner_y

    @property
    def inner_height(self) -> int:
        height = super().inner_height
        if self.frame_type:
            height -= 2
        return height

    @property
    def inner_width(self) -> int:
        width = super().inner_width
        if self.frame_type:
            width -= 2
        return width

    @property
    def inner_x(self) -> int:
        x = super().inner_x
        if self.frame_type:
            x += 1
        return x

    @property
    def inner_y(self) -> int:
        y = super().inner_y
        if self.frame_type:
            y += 1
        return y

    # Private helper methods.
    def _frame(
        self,
        frame_type: str,
        height: int,
        width: int,
        origin_y: int = 0,
        origin_x: int = 0,
        foreground: str = '',
        background: str = ''
    ) -> str:
        frame = Box(frame_type)
        result = self._get_color(foreground, background)
        result += (
            self.term.move(origin_y, origin_x)
            + frame.ltop
            + frame.top * (width - 2)
            + frame.rtop
        )
        for y in range(origin_y + 1, origin_y + height - 1):
            line = (
                self.term.move(y, origin_x) + frame.side
                + self.term.move(y, origin_x + width - 1) + frame.side
            )
            result += line
        result += (
            self.term.move(origin_y + height - 1, origin_x)
            + frame.lbot
            + frame.bot * (width - 2)
            + frame.rbot
        )
        if background or foreground:
            result += self.term.normal
        return result


class Content(Frame):
    """Create a new :class:`thurible.panel.Content` object. This class
    serves as a parent class for all panels that allow padding between
    the frame surrounding the interior of the panel and the content
    contained by the panel. The nature of that content is defined by
    the subclass. As a subclass of :class:`thurible.panel.Frame`, it
    can also take those parameters and has those public methods.

    :param content_align_h: (Optional.) The horizontal alignment
        of the contents of the panel. It defaults to center.
    :param content_align_v: (Optional.) The vertical alignment of
        the contents of the penal. It defaults to middle.
    :param content_pad_left: (Optional.) The amount of padding
        between the left inner margin of the panel and the content.
        It is measured as a float between 0.0 and 1.0, where 0.0
        is no padding and 1.0 is the entire width of the panel is
        padding. The default is 0.0.
    :param content_pad_right: (Optional.) The amount of padding
        between the right inner margin of the panel and the content.
        It is measured as a float between 0.0 and 1.0, where 0.0
        is no padding and 1.0 is the entire width of the panel is
        padding. The default is 0.0.
    :param panel_relative_width: (Optional.) The width of the
        content of the panel in comparison of the full width of
        the panel. It is a percentage expressed as a :class:`float`
        between 0.0 and 1.0, inclusive. The default is 1.0.
    :return: None.
    :rtype: NoneType
    """
    # Magic methods.
    def __init__(
        self,
        content_align_h: Optional[str] = None,
        content_align_v: str = 'middle',
        content_pad_left: Optional[float] = None,
        content_pad_right: Optional[float] = None,
        content_relative_width: Optional[float] = None,
        *args, **kwargs
    ) -> None:
        self.content_align_v = content_align_v
        self._set_content_relative_horizontal_dimensions(
            content_pad_left,
            content_relative_width,
            content_pad_right,
            content_align_h
        )
        super().__init__(*args, **kwargs)

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            super().__eq__(other)
            and self.content_align_h == other.content_align_h
            and self.content_align_v == other.content_align_v
            and self.content_pad_left == other.content_pad_left
            and self.content_pad_right == other.content_pad_right
        )

    # Properties.
    @property
    def content_width(self) -> int:
        """The width available to content within the panel after
        padding has been taken into account.

        :return: A :class:`int` object.
        :rtype: int
        """
        width = self.inner_width
        width -= self._offset_left
        width -= self._offset_right
        return width

    @property
    def content_x(self) -> int:
        """The left-most column available to content within the panel
        after padding has been taken into account.

        :return: A :class:`int` object.
        :rtype: int
        """
        x = self.inner_x
        x += self._offset_left
        return x

    @property
    def lines(self) -> list[str]:
        """The lines available to be displayed in the panel.

        :return: A :class:`list` object containing each line
            as a :class:`str`.
        :rtype: list
        """
        # This is an abstract version of the property. It will need to
        # be reimplemented for the classes that follow the Content
        # protocol.
        return ['',]

    @property
    def _offset_left(self) -> int:
        offset = super().inner_width * self.content_pad_left
        return int(offset)

    @property
    def _offset_right(self) -> int:
        offset = super().inner_width * self.content_pad_right
        return int(offset)

    # Private helper methods.
    def _align_h(self, align: str, length: int, width: int) -> int:
        """Return the amount offset the column for the horizontal
        alignment.
        """
        if align == 'left':
            x_mod = 0
        elif align == 'center':
            h_space = width - length
            x_mod = h_space // 2
        elif align == 'right':
            h_space = width - length
            x_mod = h_space
        return x_mod

    def _align_v(self, align: str, length: int, height: int) -> int:
        """Return the amount offset the column for the vertical
        alignment.
        """
        if align == 'middle':
            v_space = height - length
            y_mod = v_space // 2
        elif align == 'top':
            y_mod = 0
        elif align == 'bottom':
            v_space = height - length
            y_mod = v_space
        return y_mod

    def _set_content_relative_horizontal_dimensions(
        self,
        left: Optional[float] = None,
        width: Optional[float] = None,
        right: Optional[float] = None,
        align: Optional[str] = None,
    ) -> None:
        """Ensure the horizontal relative dimensions are set correctly."""
        # Calculate the correct values.
        left, width, right, align = self._set_relative_dimenstion(
            left,
            width,
            right,
            align,
            attr_names=(
                'content_pad_top',
                'content_relative_height',
                'content_pad_bottom',
            )
        )

        # Set the attributes.
        self.content_pad_left = left
        self.content_relative_width = width
        self.content_pad_right = right
        self.content_align_h = align


class Scroll(Content):
    """
    Create a new :class:`thurible.panel.Scroll` object. This class
    serves as a parent class for all panels that allow the user
    to scroll through content that overflows the interior of the
    panel. As a subclass of :class:`thurible.panel.Content`, it can
    also take those parameters and has those public methods.

    This class defines the following active keys:

    *   KEY_END: Scroll to the end of the content.
    *   KEY_DOWN: Scroll down in the content.
    *   KEY_HOME: Scroll to the top of the content.
    *   KEY_PGDOWN: Scroll one screen down in the content.
    *   KEY_PGUP: Scroll one page up in the content.
    *   KEY_UP: Scroll one line up in the content.

    :return: None.
    :rtype: NoneType
    """
    def __init__(
        self,
        *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)

        self._lines: list[str] = []
        self._ofbot = '[▼]'
        self._oftop = '[▲]'
        self._overflow_bottom = False
        self._overflow_top = False
        self._start = 0
        self._stop = self.inner_height
        self._wrapped_width = -1

        # Register the keyboard input the class responds to.
        self.register_key('KEY_END', self._end)
        self.register_key('KEY_DOWN', self._line_down)
        self.register_key('KEY_HOME', self._home)
        self.register_key('KEY_PGDOWN', self._page_down)
        self.register_key('KEY_PGUP', self._page_up)
        self.register_key('KEY_UP', self._line_up)

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return super().__eq__(other)

    # Property.
    @property
    def inner_height(self) -> int:
        height = super().inner_height
        if self._overflow_bottom:
            height -= 1
        if self._overflow_top:
            height -= 1
        return height

    @property
    def inner_y(self) -> int:
        y = super().inner_y
        if self._overflow_top:
            y += 1
        return y

    # Public methods.
    def action(self, key: Keystroke) -> tuple[str, str]:
        # These are the results that are returned.
        data = ''
        update = ''

        # Initial set up.
        height = self.inner_height
        width = self.inner_width
        y = self.inner_y
        x = self.inner_x
        lines = self.lines
        length = len(lines)

        # Handle input.
        if length <= height and key.name in self._active_keys:
            ...
        elif key.name in self._active_keys:
            action = self._active_keys[key.name]
            action()
            update, height, y = self._flow(
                update,
                length,
                height,
                width,
                y, x
            )
            self._overscroll(length, height)
            update += self.clear_contents()
            update += self._visible(lines, width, y, x)
        else:
            data = str(key)

        # Return the results.
        return data, update

    # Private action handlers.
    def _end(self, key: Optional[Keystroke] = None) -> str:
        """Scroll to the bottom of the content."""
        length = len(self.lines)
        self._start = length - self.inner_height
        self._stop = length
        return ''

    def _home(self, key: Optional[Keystroke] = None) -> str:
        """Scroll to the top of the content."""
        self._start = 0
        self._stop = self.inner_height
        return ''

    def _line_down(self, key: Optional[Keystroke] = None) -> str:
        """Scroll a line down in the content."""
        self._start += 1
        self._stop += 1
        return ''

    def _line_up(self, key: Optional[Keystroke] = None) -> str:
        """Scroll a line up in the content."""
        self._start -= 1
        self._stop -= 1
        return ''

    def _page_down(self, key: Optional[Keystroke] = None) -> str:
        """Scroll down one page in the content."""
        height = self.inner_height
        self._start += height
        self._stop += height
        if not self._overflow_top:
            self._start -= 1
            self._stop -= 1
        return ''

    def _page_up(self, key: Optional[Keystroke] = None) -> str:
        """Scroll up one page in the content."""
        height = self.inner_height
        self._start -= height
        self._stop -= height
        if not self._overflow_bottom:
            self._start += 1
            self._stop += 1
        return ''

    # Private helper methods.
    def _flow(
        self,
        update: str,
        length: int,
        height: int,
        width: int,
        y: int,
        x: int
    ) -> tuple[str, int, int]:
        """Manage text that overflows the visible answer."""
        if self._stop >= length - 1 and self._overflow_bottom:
            self._stop += 1
            self._overflow_bottom = False
            height += 1
            end = y + height - 1
            update += self._get_color(self.fg, self.bg)
            update += self.term.move(end, x) + ' ' * width
            if self.fg or self.bg:
                update += self.term.normal
        if self._stop < length and not self._overflow_bottom:
            self._stop -= 1
            self._overflow_bottom = True
            height -= 1
            end = y + height
            x_mod = self._align_h('center', len(self._ofbot), width)
            update += self._get_color(self.fg, self.bg)
            update += self.term.move(end, x) + ' ' * width
            update += self.term.move(end, x + x_mod) + self._ofbot
            if self.fg or self.bg:
                update += self.term.normal
        if self._start <= 1 and self._overflow_top:
            self._start -= 1
            self._overflow_top = False
            height += 1
            y -= 1
            update += self._get_color(self.fg, self.bg)
            update += self.term.move(y, x) + ' ' * width
            if self.fg or self.bg:
                update += self.term.normal
        if self._start > 0 and not self._overflow_top:
            self._start += 1
            self._overflow_top = True
            x_mod = self._align_h('center', len(self._oftop), width)
            update += self._get_color(self.fg, self.bg)
            update += self.term.move(y, x) + ' ' * width
            update += self.term.move(y, x + x_mod) + self._oftop
            if self.fg or self.bg:
                update += self.term.normal
            height -= 1
            y += 1
        return update, height, y

    def _overscroll(self, length: int, height: int) -> None:
        """Manage situations where the visible area is scrolled beyond
        the text.
        """
        if self._start < 0:
            self._start = 0
            self._stop = height
        elif self._stop > length and length > height:
            self._stop = length
            self._start = self._stop - height
        elif self._stop > length:
            self._stop = length
            self._start = 0

    def _visible(self, lines: list[str], width: int, y: int, x: int) -> str:
        """Output the lines in the display."""
        update = ''
        update += self._get_color(self.fg, self.bg)
        for i, line in enumerate(lines[self._start: self._stop]):
            x_mod = self._align_h(self.content_align_h, len(line), width)
            update += self.term.move(y + i, x + x_mod) + line
        if self.fg or self.bg:
            update += self.term.normal
        return update


class Title(Frame):
    """Create a new :class:`thurible.panel.Title` object. This class
    serves as a parent class for all panels that all the user to put
    a title on the top of the panel and a footer on the bottom of
    the frame. As a subclass of :class:`thurible.panel.Frame`, it can
    alse take those parameters and has those public methods and
    properties.

    :param footer_align: (Optional.) The horizontal alignment of the
        footer. The available options are "left", "center", and "right".
    :param footer_frame: (Optional.) Whether the frame should be capped
        on either side of the footer.
    :param footer_text: (Optional.) The text contained within the
        footer.
    :param title_align: (Optional.) The horizontal alignment of the
        title. The available options are "left", "center", and "right".
    :param title_bg: (Optional.) The background color of the title and
        footer. See the documentation for :mod:`blessed` for more detail
        on the available options.
    :param title_fg: (Optional.) The foreground color of the title and
        footer. See the documentation for :mod:`blessed` for more detail
        on the available options.
    :param title_frame: (Optional.) Whether the frame should be capped
        on either side of the title.
    :param title_text: (Optional.) The text contained within the
        title.
    :return: None.
    :rtype: NoneType
    """
    # Magic methods.
    def __init__(
        self,
        footer_align: str = 'left',
        footer_frame: bool = False,
        footer_text: str = '',
        title_align: str = 'left',
        title_bg: str = '',
        title_fg: str = '',
        title_frame: bool = False,
        title_text: str = '',
        *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.footer_text = footer_text
        self.footer_align = footer_align
        self.footer_frame = footer_frame
        self.title_text = title_text
        self.title_align = title_align
        self.title_frame = title_frame
        self.title_bg = title_bg
        self.title_fg = title_fg

        self._ofr = '[▸]'

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            super().__eq__(other)
            and self.footer_text == other.footer_text
            and self.footer_align == other.footer_align
            and self.footer_frame == other.footer_frame
            and self.title_text == other.title_text
            and self.title_align == other.title_align
            and self.title_frame == other.title_frame
            and self.title_bg == other.title_bg
            and self.title_fg == other.title_fg
        )

    def __repr__(self) -> str:
        name = self.__class__.__name__
        return f"{name}(title_text='{self.title_text}')"

    def __str__(self) -> str:
        """Return a string that will draw the entire display."""
        result = super().__str__()
        result += self.title
        result += self.footer
        return result

    # Properties.
    @property
    def footer(self) -> str:
        """The footer as a string that could be used to update the
        terminal.

        :return: A :class:`str` object.
        :rtype: str
        """
        y = self.frame_origin_y + self.frame_height - 1
        text = self.footer_text
        frame = self.footer_frame
        return self._title(text, self.footer_align, y, frame)

    @property
    def footer_frame(self) -> bool:
        return self._Title__footer_frame

    @footer_frame.setter
    def footer_frame(self, value: bool) -> None:
        if value and not self.frame_type:
            msg = 'You must set frame_type if you set footer_frame.'
            raise NoFrameTypeForFrameError(msg)
        self._Title__footer_frame = value

    @property
    def frame(self) -> str:
        result = super().frame
        if not result:
            height = self.frame_height
            width = self.frame_width
            y = self.frame_origin_y
            x = self.frame_origin_x
            bg = self.title_bg if self.title_bg else self.bg
            fg = self.title_fg if self.title_fg else self.fg
        if not result and self.title_text:
            result += self._get_color(fg, bg)
            result += self.term.move(y, x) + ' ' * width
            if bg or fg:
                result += self.term.normal
        if not result and self.footer_text:
            result += self._get_color(fg, bg)
            result += self.term.move(y + height - 1, x) + ' ' * width
            if bg or fg:
                result += self.term.normal
        return result

    @property
    def inner_height(self) -> int:
        height = super().inner_height
        if self.frame_type is None and self.title_text:
            height -= 1
        if self.frame_type is None and self.footer_text:
            height -= 1
        return height

    @property
    def inner_y(self) -> int:
        y = super().inner_y
        if self.frame_type is None and self.title:
            y += 1
        return y

    @property
    def title(self) -> str:
        """The title as a string that could be used to update the
        terminal.

        :return: A :class:`str` object.
        :rtype: str
        """
        text = self.title_text
        frame = self.title_frame
        return self._title(text, self.title_align, self.frame_origin_y, frame)

    @property
    def title_frame(self) -> bool:
        return self._Title__title_frame

    @title_frame.setter
    def title_frame(self, value: bool) -> None:
        if value and not self.frame_type:
            msg = 'You must set frame_type if you set title_frame.'
            raise NoFrameTypeForFrameError(msg)
        self._Title__title_frame = value

    # Private helper methods.
    def _title(
        self,
        title: str,
        align: str,
        y: int,
        frame: bool = False
    ) -> str:
        """Build and return a string to write the title of the panel
        to the terminal.
        """
        result = ''

        # Bail out early if there is no title.
        if not title:
            return result

        # Set up.
        x = self.inner_x
        width = self.inner_width

        # Set the color and frame for the title.
        if len(title) >= width:
            title = title[:width - 3] + self._ofr
        title = self._title_color(title)
        if isinstance(self.frame_type, str) and frame:
            title = self._title_frame(title, self.frame_type)

        # Align the title.
        if align == 'left':
            ...
        elif align == 'center':
            space = width - len(title)
            x += space // 2
        elif align == 'right':
            x += width - len(title)
        else:
            msg = f'Invalid title alignment: {self.title_align}.'
            raise InvalidTitleAlignmentError(msg)

        # Create the title and return.
        result += self.term.move(y, x) + title
        return result

    def _title_color(self, title: str) -> str:
        """Apply the title color to a title or footer."""
        bg = self.title_bg if self.title_bg else self.bg
        fg = self.title_fg if self.title_fg else self.fg
        result = self._get_color(fg, bg)
        result += title
        if bg or fg:
            result += self.term.normal
        return result

    def _title_frame(self, title: str, frame_type: str) -> str:
        """Apply the frame cap to the title or footer."""
        frame = Box(frame_type)
        bg = self.frame_bg if self.frame_bg else self.bg
        fg = self.frame_fg if self.frame_fg else self.fg
        color = self._get_color(fg, bg)
        normal = self.term.normal if color else ''
        lside = color + frame.rside + normal
        rside = color + frame.lside + normal
        return lside + title + rside



"""
util
~~~~

Miscellaneous utility functions and classes for the `thurible`
package.
"""
from queue import Queue

from typing import Optional

from blessed import Terminal


# Common values.
_term: Optional[Terminal] = None


# Common classes.
class Box:
    """Create a new :class:`thurible.util.Box` object. These objects
    track the characters used to draw a frame in a terminal. It has
    fifteen properties that return the character used for that part
    of the box:

    * top: The top
    * bot: The bottom
    * side: The sides
    * mhor: Interior horizontal lines
    * mver: Interior vertical lines
    * ltop: The top-left corner
    * mtop: Top mid-join
    * rtop: The top-right corner
    * lside: Left side mid-join
    * mid: Interior join
    * rside: Right side mid-join
    * lbot: Bottom-left corner
    * mbot: Bottom mid-join
    * rbot: Bottom-right corner

    :param kind: (Optional). Sets the set of characters used by the
        :class:`thurible.util.Box` object. It defaults to light.
        Available options include double, heavy, heavy_double_dash,
        heavy_out_light_in, heavy_quadruple_dash, light,
        light_double_dash, light_quadruple_dash, light_out_heavy_in,
        light_triple_dash.
    :param custom: (Optional). Provides a custom set of characters
        for the :class:`thurible.util.Box` object to use.
    :return: None.
    :rtype: NoneType
    """
    def __init__(
        self,
        kind: str = 'light',
        custom: Optional[str] = None
    ) -> None:
        self._names = [
            'top', 'bot', 'side',
            'mhor', 'mver',
            'ltop', 'mtop', 'rtop',
            'lside', 'mid', 'rside',
            'lbot', 'mbot', 'rbot',
        ]
        self._double = '══║═║╔╦╗╠╬╣╚╩╝'
        self._heavy = '━━┃━┃┏┳┓┣╋┫┗┻┛'
        self._heavy_double_dash = '╍╍╏╍╏' + self._heavy[3:]
        self._heavy_out_light_in = '━━┃─│┏┯┓┠┼┨┗┷┛'
        self._heavy_quadruple_dash = '┉┉┋┉┋' + self._heavy[3:]
        self._heavy_triple_dash = '┅┅┇┅┇' + self._heavy[3:]
        self._light = '──│─│┌┬┐├┼┤└┴┘'
        self._light_double_dash = '╌╌╎╌╎' + self._light[3:]
        self._light_quadruple_dash = '┈┈┊┈┊' + self._light[3:]
        self._light_out_heavy_in = '──│━┃┌┰┐┝╋┥└┸┘'
        self._light_triple_dash = '┄┄┆┄┆' + self._light[3:]
        if kind:
            self.kind = kind
        else:
            self.kind = 'light'
        if custom:
            self.custom = custom
            self.kind = 'custom'

    def __getattr__(self, name):
        try:
            index = self._names.index(name)
        except ValueError:
            return self.__dict__[name]
        return self._chars[index]

    @property
    def kind(self):
        return self._kind

    @kind.setter
    def kind(self, value):
        self._kind = value
        self._chars = getattr(self, f'_{value}')

    @property
    def custom(self):
        return self._custom

    @custom.setter
    def custom(self, value):
        strvalue = str(value)
        if len(strvalue) == 14:
            self._custom = str(strvalue)
            self._kind = 'custom'
        else:
            reason = 'The custom string must be 14 characters.'
            raise ValueError(reason)


# Common functions.
def get_queues() -> tuple[Queue, Queue]:
    """Create two :class:`queue.Queue` objects for use in communicating
    with a :func:`thurible.queued_manager` manager. This is just here
    for convenience, allowing you to use :func:`thurible.queued_manager`
    without having to import :mod:`queue`. It doesn't store the queues.

    :return: A :class:`tuple` objects, containing two
        :class:`queue.Queue` objects.
    :rtype: tuple
    """
    q_to: Queue = Queue()
    q_from: Queue = Queue()
    return q_to, q_from


def get_terminal() -> Terminal:
    """Retrieve an instance of :class:`blessed.Terminal` for use by
    :mod:`thurible` objects. Every time this is called, it will
    return the same instance, avoiding time wasting due to unnecessary
    :class:`Terminal` object initiation.

    .. note:
        Since we are using a mutable global value here, there may be
        theoretical thread safety concerns. However :mod:`thurible`
        doesn't ever change the :class:`Terminal` object. The only
        mutability is whether or not the :obj:`_term` is :obj:`None`
        or is a :class:`Terminal` object. So, I don't think thread
        safety will ever be a real issue for this. However, it may be
        worth looking into whether there are better ways to do this in
        the future. For all I know :class:`Terminal` may be a singleton,
        and this is entirely unnecessary.
    """
    global _term
    if not isinstance(_term, Terminal):
        _term = Terminal()
    return _term



"""
textdialog
~~~~~~~~~~

A text-entry dialog for terminal applications.
"""
from typing import Callable, Optional
from unicodedata import category

from blessed.keyboard import Keystroke

from thurible.dialog import Dialog
from thurible.panel import Content, Title


# Class.
class TextDialog(Content, Title):
    """Create a new :class:`thurible.TextDialog` object. This class
    displays a message to the user and allows them to input a string,
    which is send to the application. As a subclass of
    :class:`thurible.panel.Content` and :class:`thurible.panel.Title`,
    it can also take those parameters and has those public methods,
    properties, and active keys.

    This class defines the following active keys:

    *   KEY_BACKSPACE: Delete the previous character.
    *   KEY_DELETE: Delete the next character.
    *   KEY_END: Move the cursor to after the last character.
    *   KEY_HOME: Move the cursor to the first character.
    *   KEY_ENTER: Finish text entry and send input to the application.
    *   KEY_LEFT: Move the cursor to the next character.
    *   KEY_RIGHT: Move the cursor to the previous character.

    While not registered as active keys, all other key presses that
    do not result in key sequences as defined by :mod:`blessed` or
    control characters as defined by the Unicode specification are
    intercepted by the panel. The :class:`str` value of that key
    press is inserted into the text field at the position of the
    cursor.

    For more information on active keys, see :ref:`active`.

    :param message_text: The text of the prompt to be displayed to
        the user.
    :return: None.
    :rtype: NoneType
    """
    def __init__(
        self,
        message_text: str,
        *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.message_text = message_text

        self.register_key('KEY_BACKSPACE', self._delete_backwards)
        self.register_key('KEY_DELETE', self._delete)
        self.register_key('KEY_END', self._end)
        self.register_key('KEY_HOME', self._home)
        self.register_key('KEY_ENTER', self._select)
        self.register_key('KEY_LEFT', self._move_back)
        self.register_key('KEY_RIGHT', self._move_foreward)

        self._selected = 0
        self.prompt = '> '
        self.value = ''

    def __str__(self) -> str:
        result = super().__str__()

        result += self.message

        height = self.inner_height
        x = self.inner_x
        y = self._align_v('bottom', 1, height) + self.inner_y
        result += self.term.move(y, x) + self.prompt

        x += 2
        result += self.term.reverse
        result += self.term.move(y, x) + ' '
        result += self.term.normal

        return result

    # Properties
    @property
    def message(self) -> str:
        wrapped = self.term.wrap(self.message_text, width=self.inner_width)
        length = len(wrapped)
        y = self._align_v('middle', length, self.inner_height) + self.inner_y
        x = self.inner_x
        result = ''
        for i, line in enumerate(wrapped):
            result += f'{self.term.move(y + i, x)}{line}'
        return result

    # Public methods.
    def action(self, key: Keystroke) -> tuple[str, str]:
        # If, somehow, we received something that isn't a keystroke,
        # something has gone seriously wrong.
        if not isinstance(key, Keystroke):
            cls_name = type(key).__name__
            msg = f'Can only accept Keystrokes. Received: {cls_name}.'
            raise ValueError(msg)

        # These are the results that are returned.
        data = ''
        update = ''

        # Handle the keys with defined behavior.
        if repr(key) in self._active_keys:
            handler = self._active_keys[repr(key)]
            data = handler(key)

        # If it's non-printable and has no defined behavior, pass it
        # back to the program to figure out.
        elif key.is_sequence or category(str(key)) == 'Cc':
            data = str(key)

        # If it's printable and the cursor is in the middle of the
        # text being typed, insert the character in front of the
        # current position.
        elif self._selected < len(self.value):
            index = self._selected
            self.value = (self.value[0:index] + str(key) + self.value[index:])
            self._selected += 1

        # Otherwise, add it to the end of the text being typed.
        else:
            self.value += str(key)
            self._selected += 1

        # If data isn't being returned, we probably need to update the
        # terminal to show what happened.
        if not data:
            # Set up.
            prompt_length = len(self.prompt)
            height = self.inner_height
            width = self.inner_width - prompt_length
            x = self.inner_x + prompt_length
            y = self._align_v('bottom', 1, height) + self.inner_y

            # Create the string used to update the terminal.
            update += self.term.move(y, x) + f'{self.value:<{width}}'
            update += self.term.reverse
            update += self.term.move(y, x + self._selected)
            if self._selected < len(self.value):
                selected_char = self.value[self._selected]
            else:
                selected_char = ' '
            update += selected_char
            update += self.term.normal

        # Return the results as a tuple.
        return data, update

    # Private action handlers.
    def _delete(self, key: Optional[Keystroke]) -> str:
        """Delete the selected character."""
        index = self._selected
        self.value = self.value[:index] + self.value[index + 1:]
        return ''

    def _delete_backwards(self, key: Optional[Keystroke]) -> str:
        """Delete the previous character."""
        self._selected -= 1
        return self._delete(key)

    def _end(self, key: Optional[Keystroke]) -> str:
        """Move the cursor to the last position."""
        self._selected = len(self.value)
        return ''

    def _home(self, key: Optional[Keystroke]) -> str:
        """Move the cursor to the first character."""
        self._selected = 0
        return ''

    def _move_back(self, key: Optional[Keystroke]) -> str:
        """Move the cursor back one character."""
        if self._selected > 0:
            self._selected -= 1
        return ''

    def _move_foreward(self, key: Optional[Keystroke]) -> str:
        """Move the cursor foreward one character."""
        if self._selected < len(self.value):
            self._selected += 1
        return ''

    def _select(self, key: Optional[Keystroke] = None) -> str:
        """Return the name of the selected option."""
        return self.value



"""
eventmanager
~~~~~~~~~~~~

A manager that uses events sent from the user interface to drive
application flow.
"""
from queue import Queue
from threading import Thread
from typing import Callable, Optional

from thurible.panel import Panel
from thurible.thurible import queued_manager
from thurible.util import get_queues
import thurible.messages as tm


# Manager.
def event_manager(
    event_map: Optional[dict[type, Callable]] = None,
    initial_panel: Optional[Panel] = None
) -> None:
    """Manage a terminal display by mapping :ref:`response-messages`
    to functions in your application.

    :param event_map: (Optional.) A :class:`dict` mapping the
        :class:`thurible.panel.Message` types managers can send to
        applications (:ref:`response-messages`) to functions in
        your application. These functions must accept a
        :class:`queue.Queue` object and the response message as
        parameters. It must return a :class:`bool` indicating
        whether the application should continue running.
    :param initial_panel: (Optional.) The first panel displayed in
        the terminal. While this is technically optional, that's
        just for testing purposes. You should really provide
        this to the manager. The panel passed this way will be
        stored as "__init".
    :return: None
    :rtype: NoneType

    Usage::

        from thurible import event_manager, Splash
        import thurible.messages as tm

        # Create the event handlers.
        def data_handler(msg, q_to):
            msg = tm.End('Quitting.')
            q_to.put(msg)
            return False

        def ending_handler(msg, q_to):
            if msg.exception:
                raise msg.exception
            return False

        # Map the handlers to event messages.
        event_map = {
            tm.Data: data_handler,
            tm.Ending: ending_handler,
        }

        # Create the panel to display when the manager starts.
        splash = Splash('SPAM!')

        # Run the event_manager.
        event_manager(event_map, splash)
    """
    if not event_map:
        event_map = {}

    q_to, q_from = get_queues()
    T = Thread(target=queued_manager, args=(q_to, q_from))
    run = True

    try:
        T.start()
        if initial_panel:
            q_to.put(tm.Store('__init', initial_panel))
            q_to.put(tm.Show('__init'))

        while run:
            run = _check_for_message(q_to, q_from, event_map)

    except KeyboardInterrupt as ex:
        reason = 'Keyboard Interrupt'
        msg = tm.End(reason)
        q_to.put(msg)
        raise ex


# Private functions.
def _check_for_message(
    q_to: Queue,
    q_from: Queue,
    event_map: dict[type, Callable]
) -> bool:
    """Check for and handle UI messages."""
    run = True
    if not q_from.empty():
        msg = q_from.get()

        for msg_type in event_map:
            if isinstance(msg, msg_type):
                run = event_map[msg_type](msg, q_to)
                break
        else:
            if isinstance(msg, tm.Ending):
                run = False
    return run



"""
splash
~~~~~~

A splash screen for terminal applications.
"""
from blessed import Terminal

from thurible.panel import Content, Title


class Splash(Content, Title):
    """Create a new :class:`thurible.Splash` object. This class
    creates a splash screen that can be displayed in the terminal.
    As a subclass of :class:`thurible.panel.Content` and
    :class:`thurible.panel.Title`, it can also take those parameters
    and has those public methods and properties.

    :param content: (Optional.) The text to display within the
        interior of the panel.
    :return: None.
    :rtype: NoneType
    """
    # Magic methods.
    def __init__(
        self,
        content: str = '',
        *args, **kwargs
    ) -> None:
        self.content = content
        super().__init__(*args, **kwargs)

    def __str__(self) -> str:
        """Return a string that will draw the entire panel."""
        # Set up.
        lines = self.lines
        height = self.inner_height
        width = self.inner_width
        y = self.inner_y
        x = self.inner_x
        result = super().__str__()

        length_v = len(lines)
        y += self._align_v(self.content_align_v, length_v, height)

        for i, line in enumerate(lines):
            length_h = len(line)
            x_mod = self._align_h(self.content_align_h, length_h, width)
            result += self.term.move(y + i, x + x_mod) + line
        return result

    # Properties.
    @property
    def lines(self) -> list[str]:
        """The lines of text available to be displayed in the panel
        after they have been wrapped to fit the width of the
        interior of the panel.

        :return: A :class:`list` object containing each line of
            text as a :class:`str`.
        :rtype: list
        """
        return self.content.split('\n')



"""
messages
~~~~~~~~

Classes for communicating with `thurible` managers.
"""
from dataclasses import dataclass
from typing import Optional, Sequence

from thurible.dialog import cont
from thurible.menu import Option
from thurible.panel import Message, Panel


# Command messages.
@dataclass
class Alert(Message):
    """Create a new :class:`thurible.messages.Alert` object. This
    object is a command message used to instruct a manager to
    show an alert message to the user.

    :param name: (Optional.) The name the manager will use to store
        the :class:`thurible.Dialog` object created in response to
        this message. The default name is "alert".
    :param title: (Optional.) The title of the alert.
    :param text: (Optional.) The text of the alert. The default value
        is "Error."
    :param options: (Optional.) The options given to the user for
        responding to the alert. The default is "Continue".
    :return: None.
    :rtype: NoneType
    """
    name: str = 'alert'
    title: str = ''
    text: str = 'Error.'
    options: Sequence[Option] = cont


@dataclass
class Delete(Message):
    """Create a new :class:`thurible.messages.Delete` object. This
    object is a command message used to instruct a manager to
    delete a stored panel.

    :param name: The name of the panel to delete.
    :return: None.
    :rtype: NoneType
    """
    name: str


@dataclass
class Dismiss(Message):
    """Create a new :class:`thurible.messages.Dismiss` object. This
    object is a command message used to stop displaying an alert.

    :param name: (Optional.) The name of the panel to dismiss.
    :return: None.
    :rtype: NoneType
    """
    name: str = 'alert'


@dataclass
class End(Message):
    """Create a new :class:`thurible.messages.End` object. This
    object is a command message used to instruct a manager to
    end the manager loop and quit.

    :param text: (Optional.) A message to print for the user after
        the manager loop ends.
    :return: None.
    :rtype: NoneType
    """
    text: str = ''


@dataclass
class Ping(Message):
    """Create a new :class:`thurible.messages.Ping` object. This
    object is a command message used to instruct a manager to
    reply with a :class:`thurible.message.Pong` message, proving
    the manager is still listening for and responding to messages.

    :param name: A unique name used to identify the resulting
        :class:`thurible.message.Pong` message as being caused
        by this message.
    :return: None.
    :rtype: NoneType
    """
    name: str


@dataclass
class Show(Message):
    """Create a new :class:`thurible.messages.Show` object. This
    object is a command message used to instruct a manager to
    display a stored panel.

    :param name: The name of the panel to display.
    :return: None.
    :rtype: NoneType
    """
    name: str


@dataclass
class Showing(Message):
    """Create a new :class:`thurible.messages.Showing` object. This
    object is a command message used to instruct a manager to
    respond with a :class:`thurible.messages.Shown` message
    with the name of the currently displayed panel.

    :param name: (Optional.) A unique name used to identify the
        resulting :class:`thurible.message.Shown` message as being
        caused by this message.
    :return: None.
    :rtype: NoneType
    """
    name: str = ''


@dataclass
class Store(Message):
    """Create a new :class:`thurible.messages.Store` object. This
    object is a command message used to instruct a manager to
    store a panel for later display.

    :param name: The name of the panel to store.
    :param name: The panel to store.
    :return: None.
    :rtype: NoneType
    """
    name: str
    display: Panel


@dataclass
class Storing(Message):
    """Create a new :class:`thurible.messages.Storing` object. This
    object is a command message used to instruct a manager to
    respond with a :class:`thurible.message.Stored` object
    containing the names of the currently stored panels.

    :param name: (Optional.) A unique name used to identify the
        resulting :class:`thurible.message.Stored` message as being
        caused by this message.
    :return: None.
    :rtype: NoneType
    """
    name: str = ''


# Response messages.
@dataclass
class Data(Message):
    """Create a new :class:`thurible.messages.Data` object. This
    object is a response message used to send data back to the
    application.

    :param value: The data being sent to the application.
    :return: None.
    :rtype: NoneType
    """
    value: str


@dataclass
class Ending(Message):
    """Create a new :class:`thurible.messages.Ending` object. This
    object is a response message used to inform the application
    that the manager is ending.

    :param reason: (Optional.) The reason the manager loop is
        ending.
    :param ex: (Optional.) The exception causing the manager
        loop to end.
    :return: None.
    :rtype: NoneType
    """
    reason: str = ''
    ex: Optional[Exception] = None


@dataclass
class Pong(Message):
    """Create a new :class:`thurible.messages.Pong` object. This
    object is a response message used to respond to a
    :class:`thurible.messages.Ping` message.

    :param name: The name of the :class:`thurible.messages.Ping`
        message that caused this response.
    :return: None.
    :rtype: NoneType
    """
    name: str


@dataclass
class Shown(Message):
    """Create a new :class:`thurible.messages.Shown` object. This
    object is a response message used to respond to a
    :class:`thurible.messages.Showing` message.

    :param name: The name of the :class:`thurible.messages.Showing`
        message that caused this response.
    :param display: The name of the panel being displayed when the
        :class:`thurible.messages.Showing` was received.
    :return: None.
    :rtype: NoneType
    """
    name: str
    display: str


@dataclass
class Stored(Message):
    """Create a new :class:`thurible.messages.Stored` object. This
    object is a response message used to respond to a
    :class:`thurible.messages.Storing` message.

    :param name: The name of the :class:`thurible.messages.Storing`
        message that caused this response.
    :param display: The names of the panel being stored when the
        :class:`thurible.messages.Storing` message was received.
    :return: None.
    :rtype: NoneType
    """
    name: str
    stored: tuple[str, ...]



"""
text
~~~~

An object for displaying a text area in a terminal.
"""
from dataclasses import dataclass
from typing import Optional

from blessed import Terminal
from blessed.keyboard import Keystroke

from thurible.panel import Scroll, Title
from thurible.util import Box


# Utility classes.
@dataclass
class Option:
    """A command or menu option.

    :param name: The name of the option.
    :param hotkey: (Optional.) A hotkey that can be used to invoke
        the option.
    """
    name: str
    hotkey: Optional[str] = None


# General classes.
class Menu(Scroll, Title):
    """Create a new :class:`thurible.Menu` object. This class provides
    a list of options the user can select. As a subclass of
    :class:`thurible.panel.Scroll` and :class:`thurible.panel.Title`,
    it can also take those parameters and has those public methods,
    properties, and active keys.

    :class:`thurible.Menu` adds the additional active key:

    *   KEY_ENTER: Select the highlighted option.
    *   Optional hot keys to highlight the options, as defined in the
        :class:`thurible.Option` object for the option.

    :class:`thurible.Menu` modifies the behavior of the following
    active keys:

    *   KEY_END: Highlight the last option, scrolling if needed.
    *   KEY_DOWN: Highlight the next option, scrolling if needed.
    *   KEY_HOME: Highlight the first option, scrolling if needed.
    *   KEY_PGDOWN: Scroll to and highlight the option one screen down.
    *   KEY_PGUP: Scroll to and highlight the option one screen up.
    *   KEY_UP: Highlight the previous option, scrolling if needed.

    For more information on active keys, see :ref:`active`.

    :param options: A sequence of :class:`thurible.Option` objects
        defining the options available to the user.
    :param option_align_h: (Optional.) The horizontal alignment
        of the options within the area that would be highlighted when
        the option is highlighted. If you think of each option as a
        button, it's how the text is aligned on the face of the button.
        It defaults to "left".
    :param select_bg: (Optional.) The background color used to
        highlight an option.
    :param select_fg: (Optional.) The foreground color used to
        highlight an option.
    :param content_align_h: (Optional.) The horizontal alignment
        of the contents of the panel. It defaults to "left". See
        :class:`thurible.panel.Content` for more information.
    :param content_align_v: (Optional.) The vertical alignment
        of the contents of the panel. It defaults to "top".
    :return: None.
    :rtype: NoneType
    """
    # Magic methods.
    def __init__(
        self,
        options: tuple[Option],
        option_align_h: str = 'left',
        select_bg: str = '',
        select_fg: str = '',
        content_align_v: str = 'top',
        *args, **kwargs
    ) -> None:
        self.options = options
        self.option_align_h = option_align_h
        self.select_bg = select_bg
        self.select_fg = select_fg
        if (
            'content_pad_left' not in kwargs
            and 'conent_pad_right' not in kwargs
            and 'content_align_h' not in kwargs
        ):
            kwargs['content_align_h'] = 'left'
        kwargs['content_align_v'] = content_align_v
        super().__init__(*args, **kwargs)

        self._selected = 0
        self._active_keys['KEY_ENTER'] = self._select
        for option in self.options:
            hotkey = f"'{option.hotkey}'"
            self._active_keys[hotkey] = self._hotkey

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            super().__eq__(other)
            and self.options == other.options
            and self.select_bg == other.select_bg
            and self.select_fg == other.select_fg
        )

    def __str__(self) -> str:
        """Return a string that will draw the entire panel."""
        # Set up.
        lines = self.lines
        length = len(lines)
        height = self.inner_height
        width = self.content_width
        y = self.inner_y
        x = self.content_x
        self._start = 0
        self._stop = height
        result = super().__str__()

        # Create the display string and return.
        y += self._align_v(self.content_align_v, length, height)
        result, height, y = self._flow(result, length, height, width, y, x)
        self._overscroll(length, height)
        result += self._visible(lines, width, y, x)
        return result

    # Properties.
    @property
    def field_width(self) -> int:
        """The width of the highlightable area of each option as
        determined by the option with the most characters.

        :return: A :class:`int` object.
        :rtype: int
        """
        return max(len(opt.name) for opt in self.options)

    @property
    def lines(self) -> list[str]:
        """A :class:`list` of :class:`str` objects used to display
        the panel in the terminal.

        :return: A :class:`list` object of :class:`str` objects.
        :rtype: list
        """
        lines = []
        fwidth = self.field_width
        align = '<'
        if self.option_align_h == 'center':
            align = '^'
        if self.option_align_h == 'right':
            align = '>'

        for option in self.options:
            line = f'{option.name:{align}{fwidth}}'
            lines.append(line)

        return lines

    # Public methods.
    def action(self, key: Keystroke) -> tuple[str, str]:
        # These are the results that are returned.
        data = ''
        update = ''

        # Initial set up.
        height = self.inner_height
        width = self.content_width
        y = self.inner_y
        x = self.content_x
        lines = self.lines
        length = len(lines)

        # Handle input.
        if repr(key) in self._active_keys:
            handler = self._active_keys[repr(key)]
            data = handler(key)
        else:
            data = str(key)

        # Create any needed updates to the terminal.
        if not data:
            update, height, y = self._flow(
                update,
                length,
                height,
                width,
                y, x
            )
            self._overscroll(length, height)
            update += self._visible(lines, width, y, x)

        # Return the results.
        return data, update

    # Private action handlers.
    def _end(self, key: Optional[Keystroke] = None) -> str:
        """Select the last option and scroll to it."""
        length = len(self.lines)
        self._selected = length - 1
        self._start = length - self.inner_height
        self._stop = length
        return ''

    def _home(self, key: Optional[Keystroke] = None) -> str:
        """Select the first option and scroll to it."""
        self._selected = 0
        self._start = 0
        self._stop = self.inner_height
        return ''

    def _hotkey(self, key: Optional[Keystroke] = None) -> str:
        hotkeys = [option.hotkey for option in self.options]
        length = len(self.options)
        height = self.inner_height
        self._selected = hotkeys.index(str(key))
        if self._start > self._selected:
            self._start = self._selected
            self._stop = self._start + height
        elif self._stop <= self._selected:
            self._stop = self._selected
            self._start = self._stop - height
            if length > height and self._selected == length - 2:
                self._start += 1
                self._stop += 1
        return ''

    def _line_down(self, key: Optional[Keystroke] = None) -> str:
        """Select the next option."""
        if self._selected < len(self.options) - 1:
            self._selected += 1
        if self._selected >= self._stop:
            self._start += 1
            self._stop += 1
        return ''

    def _line_up(self, key: Optional[Keystroke] = None) -> str:
        """Select the pervious option."""
        if self._selected > 0:
            self._selected -= 1
        if self._selected < self._start:
            self._start -= 1
            self._stop -= 1
        return ''

    def _page_down(self, key: Optional[Keystroke] = None) -> str:
        """Scroll down one page in the content."""
        height = self.inner_height
        self._selected += height
        self._start += height
        self._stop += height
        if not self._overflow_top:
            self._start -= 1
            self._stop -= 1
        return ''

    def _page_up(self, key: Optional[Keystroke] = None) -> str:
        """Scroll up one page in the content."""
        height = self.inner_height
        self._selected -= height
        self._start -= height
        self._stop -= height
        if not self._overflow_bottom:
            self._start += 1
            self._stop += 1
        return ''

    def _select(self, key: Optional[Keystroke] = None) -> str:
        """Return the name of the selected option."""
        return self.options[self._selected].name

    # Private helper methods.
    def _color_selection(self) -> str:
        result: str = self.term.reverse
        if self.select_bg or self.select_fg:
            result = self._get_color(self.select_fg, self.select_bg)
        return result

    def _overscroll(self, length: int, height: int) -> None:
        if self._selected < 0:
            self._selected = 0
        elif self._selected >= length:
            self._selected = length - 1
        super()._overscroll(length, height)

    def _visible(self, lines: list[str], width: int, y: int, x: int) -> str:
        """Output the lines in the display."""
        # Set the base colors for the menu options.
        update = self._get_color(self.fg, self.bg)

        # Create the visible options.
        for i, line in enumerate(lines[self._start: self._stop]):
            x_mod = self._align_h(self.content_align_h, len(line), width)

            # Use the selection colors if the option is selected. The
            # `opt_index` variable here translates working row in the
            # terminal display to the index of the items in the list
            # of menu options. It's needed because we aren't tracking
            # the selection in the options list itself.
            opt_index = i + self._start
            if opt_index == self._selected:
                update += self._color_selection()

            # Create the option.
            update += self.term.move(y + i, x + x_mod) + line

            # Revert to the base colors if this option was selected.
            if opt_index == self._selected and not (self.fg or self.bg):
                update += self.term.normal
            elif opt_index == self._selected:
                update += self.term.normal
                update += self._get_color(self.fg, self.bg)

        # End the coloring so we don't have to worry about it the next
        # time we print a string, and return the options.
        if self.fg or self.bg:
            update += self.term.normal
        return update



"""
text
~~~~

An object for displaying a data table in a terminal.
"""
from dataclasses import astuple, dataclass, fields
import unicodedata as ucd
from typing import Any, Optional, Protocol, Sequence

from blessed import Terminal
from blessed.keyboard import Keystroke

from thurible.panel import Scroll, Title
from thurible.util import Box as Frame


# Protocols for type hinting.
class Dataclass(Protocol):
    """A class definition that can be used to represent dataclasses
    for type hinting.
    """
    __dataclass_fields__: dict


# Classes.
class Table(Scroll, Title):
    """Create a new :class:`thurible.Table` object. This class displays
    a table of data to the user. As a subclass of
    :class:`thurible.panel.Scroll` and :class:`thurible.panel.Title`,
    it can also take those parameters and has those public methods,
    properties, and active keys.

    :param records: A sequence of dataclasses that will be displayed
        within the panel. The data held by the dataclass can be of
        any type, but it must be able to be coerced into a :class:str.
    :param inner_frame: (Optional.) Whether there should be a visible
        frame around each cell in the panel.
    :param content_align_h: (Optional.) The horizontal alignment
        of the contents of the panel. It defaults to "left".
    :param content_align_v: (Optional.) The vertical alignment
        of the contents of the panel. It defaults to "top".
    :return: None.
    :rtype: NoneType
    """
    def __init__(
        self,
        records: Sequence[Dataclass],
        inner_frame: bool = False,
        content_align_h: str = 'left',
        content_align_v: str = 'top',
        *args, **kwargs
    ) -> None:
        self.records = records
        self.inner_frame = inner_frame
        kwargs['content_align_h'] = content_align_h
        kwargs['content_align_v'] = content_align_v
        super().__init__(*args, **kwargs)

        self._char_true = '█'
        self._char_false = '▁'
#         self._ofr = '[▶︎]'

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            super().__eq__(other)
            and self.records == other.records
            and self.inner_frame == other.inner_frame
            and self._char_true == other._char_true
            and self._char_false == other._char_false
            and self._ofr == other._ofr
        )

    def __str__(self) -> str:
        """Return a string that will draw the entire panel."""
        # Set up.
        lines = self.lines
        length = len(lines)
        height = self.inner_height
        width = self.inner_width
        y = self.inner_y
        x = self.inner_x
        self._start = 0
        self._stop = height
        result = super().__str__()

        # Create the display string and return.
        y += self._align_v(self.content_align_v, length, height)
        result, height, y = self._flow(result, length, height, width, y, x)
        self._overscroll(length, height)
        result += self._visible(lines, width, y, x)
        return result

    # Properties.
    @property
    def field_names(self) -> list[str]:
        """The names of each field of data contained within the records
        being displayed.

        :return: A :class:`list` object containing each name as
            a :class:`str` object.
        :rtype: list
        """
        if '_field_names' not in self.__dict__:
            self._field_names = [f.name for f in fields(self.records[0])]
        return self._field_names

    @property
    def field_widths(self) -> list[int]:
        """The width in characters of each field in the table, as
        determined by the longest value for this field found in
        the dataclasses.

        :return: A :class:`list` object containing each width as an
        :class:`int`.
        :rtype: list
        """
        if '_field_widths' not in self.__dict__:
            fnames = self.field_names
            fwidths = self._calc_field_widths(self.records, fnames)
            if sum(fwidths) + len(fwidths) - 1 > self.inner_width:
                fwidths = self._calc_field_widths_overflow(fwidths)
            self._field_widths = fwidths
        return self._field_widths

    @property
    def lines(self) -> list[str]:
        """The lines of text available to be displayed in the panel
        after they have been wrapped to fit the width of the
        interior of the panel.

        :return: A :class:`list` object containing each line of
            text as a :class:`str`.
        :rtype: list
        """
        lines = []

        align = '<'
        if self.content_align_h == 'center':
            align = '^'
        elif self.content_align_h == 'right':
            align = '>'
        width = self.inner_width
        fnames = self.field_names
        fwidths = self.field_widths
        if self.frame_type:
            frame = Frame(self.frame_type)

        for record in self.records:
            line = ''
            seperator = ' '
            if self.inner_frame and self.frame_type:
                seperator = frame.mver
            for fname, fwidth in zip(fnames, fwidths):
                line += self._get_field_value(record, fname, fwidth)
                if fname != fnames[-1]:
                    line += seperator
            line = f'{line:{align}{width}}'
            if self.frame_type:
                line = frame.mver + line + frame.mver
            lines.append(line)

            if self.inner_frame and self.frame_type:
                line = ''
                seperator = frame.mid
                for fname, fwidth in zip(fnames, fwidths):
                    line += frame.mhor * fwidth
                    if fname != fnames[-1]:
                        line += seperator
                line = f'{line:{frame.mhor}{align}{width}}'
                line = frame.lside + line + frame.rside
                lines.append(line)

        if self.inner_frame and self.frame_type:
            lines = lines[:-1]
        return lines

    # Private helper methods.
    def _calc_field_widths(
        self,
        records: Sequence[Dataclass],
        names: list[str],
    ) -> list[int]:
        result = []
        for name in names:
            data = [getattr(record, name) for record in records]
            width = 1
            if not isinstance(data[0], bool):
                width = self._calc_max_width(data)
            result.append(width)
        return result

    def _calc_field_widths_overflow(self, fwidths: list[int]) -> list[int]:
        # We don't want to affect the small fields, so calculate the
        # size of an average field to use to determine which fields are
        # small.
        num_fields = len(fwidths)
        divs = num_fields - 1
        avail = self.inner_width - divs
        avg_fwidth = avail // num_fields

        # Determine which fields are big.
        big_mask = [w > avg_fwidth for w in fwidths]

        # Determine the new field size for the big fields.
        big_avail = avail - sum(w for w in fwidths if w <= avg_fwidth)
        num_big = len([item for item in big_mask if item])
        big_avg = big_avail // num_big

        # Set the big fields to their new sizes.
        result = fwidths[:]
        for i in range(num_fields):
            if big_mask[i]:
                result[i] = big_avg

        # Add back in any extra space left over from using floor division
        # to calculate the new big field size.
        remain = big_avail % num_big
        bigs = [i for i, v in enumerate(big_mask) if v]
        for i in bigs[:remain]:
            result[i] += 1

        # Return the result.
        return result

    def _calc_max_width(self, data: Sequence[Any]) -> int:
        """Determine the length of the longest string."""
        strings = (str(datum) for datum in data)
        return max(len(s) for s in strings)

    def _frame(
        self,
        frame_type: str,
        height: int,
        width: int,
        origin_y: int = 0,
        origin_x: int = 0,
        foreground: str = '',
        background: str = ''
    ) -> str:
        frame = Frame(frame_type)
        result = super()._frame(
            frame_type,
            height,
            width,
            origin_y,
            origin_x,
            foreground,
            background
        )
        result += self._get_color(foreground, background)

        if self.inner_frame and self.frame_type:
            fwidths = self.field_widths[:-1]
            x = origin_x + 1
            bottom_y = origin_y + height - 1
            for fwidth in fwidths:
                x += fwidth
                result += self.term.move(origin_y, x) + frame.mtop
                result += self.term.move(bottom_y, x) + frame.mbot
                x += 1

        if background or foreground:
            result += self.term.normal
        return result

    def _get_field_value(
        self,
        record: Dataclass,
        field: str,
        width: int
    ) -> str:
        align = '<'
        value = getattr(record, field)

        if isinstance(value, bool) and value:
            value = self._char_true
        elif isinstance(value, bool):
            value = self._char_false
        elif value and isinstance(value, int):
            align = '>'

        value = str(value)
        if len(value) > width:
            value = value[:width - 3] + self._ofr

        for i, char in enumerate(value):
            if ucd.category(char) == 'Cc':
                value = value[0:i] + ' ' + value[i + 1:]

        return f'{value:{align}{width}}'

    def _visible(self, lines: list[str], width: int, y: int, x: int) -> str:
        """Output the lines in the display."""
        update = ''
        update += self._get_color(self.fg, self.bg)
        for i, line in enumerate(lines[self._start: self._stop]):
            x_mod = self._align_h(self.content_align_h, len(line), width)
            if self.frame_type:
                x_mod -= 1
            update += self.term.move(y + i, x + x_mod) + line
        if self.fg or self.bg:
            update += self.term.normal
        return update



"""
text
~~~~

An object for displaying a text area in a terminal.
"""
from blessed import Terminal
from blessed.keyboard import Keystroke

from thurible.panel import Scroll, Title
from thurible.util import Box


class Text(Scroll, Title):
    """Create a new :class:`thurible.Text` object. This class displays
    text to the document and allows the user to scroll through that
    text if it is too long to fit in the terminal window. As a subclass
    of :class:`thurible.panel.Scroll` and :class:`thurible.panel.Title`,
    it can also take those parameters and has those public methods,
    properties, and active keys.

    :param content: (Optional.) The text to display in the interior
        of the panel.
    :param content_align_h: (Optional.) The horizontal alignment
        of the contents of the panel. It defaults to "left".
    :param content_align_v: (Optional.) The vertical alignment
        of the contents of the panel. It defaults to "top".
    :return: None.
    :rtype: NoneType
    """
    # Magic methods.
    def __init__(
        self,
        content: str = '',
        content_align_h: str = 'left',
        content_align_v: str = 'top',
        *args, **kwargs
    ) -> None:
        self.content = content
        kwargs['content_align_h'] = content_align_h
        kwargs['content_align_v'] = content_align_v
        super().__init__(*args, **kwargs)

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            super().__eq__(other)
            and self.content == other.content
        )

    def __str__(self) -> str:
        """Return a string that will draw the entire panel."""
        # Set up.
        lines = self.lines
        length = len(lines)
        height = self.inner_height
        width = self.inner_width
        y = self.inner_y
        x = self.inner_x
        self._start = 0
        self._stop = height
        result = super().__str__()

        # Create the display string and return.
        y += self._align_v(self.content_align_v, length, height)
        result, height, y = self._flow(result, length, height, width, y, x)
        self._overscroll(length, height)
        result += self._visible(lines, width, y, x)
        return result

    # Properties.
    @property
    def lines(self) -> list[str]:
        """The lines of text available to be displayed in the panel
        after they have been wrapped to fit the width of the
        interior of the panel.

        :return: A :class:`list` object containing each line of
            text as a :class:`str`.
        :rtype: list
        """
        width = self.inner_width
        if width != self._wrapped_width:
            wrapped = self.term.wrap(self.content, width)
            self._lines = wrapped
            self._wrapped_width = width
        return self._lines



"""
progress
~~~~~~~~

An object for announcing the progress towards a goal.
"""
from collections import deque
from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Sequence

from thurible.panel import Content, Message, Title


# Message classes.
@dataclass
class NoTick(Message):
    """Create a new :class:`thurible.progress.NoTick` object. When
    sent to :meth:`thurible.Progress.update`, this will not cause
    the progress bar to advance.

    :param message: A message to display.
    :return: None.
    :rtype: NoneType
    """
    message: str = ''


@dataclass
class Tick(Message):
    """Create a new :class:`thurible.progress.Tick` object. When
    sent to :meth:`thurible.Progress.update`, this will cause the
    progress bar to advance.

    :param message: A message to display.
    :return: None.
    :rtype: NoneType
    """
    message: str = ''


# Panel class.
class Progress(Content, Title):
    """Create a new :class:`thurible.Progress` object. This
    object displays a bar representing how much progress has
    been achieved towards a goal. As a subclass of
    :class:`thurible.panel.Content` and :class:`thurible.panel.Title`,
    it can also take those parameters and has those public methods
    and properties.

    :param steps: The number of steps required to achieve the
        goal.
    :param progress: (Optional.) The number of steps that have been
        completed.
    :param bar_bg: (Optional.) A string describing the background
        color of the bar. See the documentation for :mod:`blessed`
        for more detail on the available options.
    :param bar_fg: (Optional.) A string describing the foreground
        color of the bar. See the documentation for :mod:`blessed`
        for more detail on the available options.
    :param max_messages: (Optional.) How many status messages should
        be stored to be displayed.
    :param messages: (Optional.) Any status messages to start in the
        display. Since new messages are added to the display at the
        top, the messages passed in this sequence should be stored
        in reverse chronological order.
    :param timestamp: (Optional.) Add a timestamp to the messages
        when they are displayed.
    :return: None.
    :rtype: NoneType
    """
    def __init__(
        self,
        steps: int,
        progress: int = 0,
        bar_bg: str = '',
        bar_fg: str = '',
        max_messages: int = 0,
        messages: Optional[Sequence[str]] = None,
        timestamp: bool = False,
        *args, **kwargs
    ) -> None:
        self._notick = False
        self._t0 = datetime.now()
        self._wrapped_width = -1

        self.steps = steps
        self.progress = progress
        self.bar_bg = bar_bg
        self.bar_fg = bar_fg
        self.max_messages = max_messages
        self.timestamp = timestamp
        self.messages: deque = deque(maxlen=self.max_messages)
        if messages:
            for msg in messages:
                self._add_message(msg)
        super().__init__(*args, **kwargs)

    def __str__(self) -> str:
        """Return a string that will draw the entire panel."""
        # Set up.
        result = super().__str__()
        height = 1 + self.max_messages
        y = self._align_v('middle', height, self.inner_height) + self.inner_y
        x = self.content_x

        # Add the progress bar.
        result += self.term.move(y, x) + self.progress_bar
        y += 1

        # Add messages.
        if self.max_messages:
            result += self._visible_messages(x, y)

        # Return the resulting string.
        return result

    # Properties.
    @property
    def lines(self) -> list[str]:
        """The lines of text available to be displayed in the panel
        after they have been wrapped to fit the width of the
        interior of the panel. A message from the application may
        be split into multiple lines.

        :return: A :class:`list` object containing each line of
            text as a :class:`str`.
        :rtype: list
        """
        width = self.content_width
        if width != self._wrapped_width:
            wrapped = []
            for line in self.messages:
                wrapped.extend(self.term.wrap(line, width=width))
            self._lines = wrapped
            self._wrapped_width = width
        return self._lines

    @property
    def progress_bar(self) -> str:
        """The progress bar as a string.

        :return: A :class:`str` object.
        :rtype: str
        """
        # Color the bar.
        result = self._get_color(self.bar_fg, self.bar_bg)

        # Unicode has characters to fill eighths of a character,
        # so we can resolve progress at eight times the width available
        # to us.
        notches = self.content_width * 8

        # Determine the number of notches filled.
        notches_per_step = notches / self.steps
        progress_notches = notches_per_step * self.progress
        full = int(progress_notches // 8)
        part = int(progress_notches % 8)

        # The Unicode characters we are using are the block fill
        # characters in the range 0x2588–0x258F. This takes
        # advantage of the fact they are in order to make it
        # easier to find the one we need.
        blocks = {i: chr(0x2590 - i) for i in range(1, 9)}

        # Build the bar.
        progress = blocks[8] * full
        if part:
            progress += blocks[part]
        result += f'{progress:<{self.content_width}}'

        # If a color was set, return to normal to avoid unexpected
        # behavior. Then return the string.
        if self.bar_bg or self.bar_fg:
            result += self.term.normal
        return result

    # Public methods.
    def update(self, msg: Message) -> str:
        """Act on a message sent by the application.

        :class:`thurible.Progress` responds to the following
        update messages:

        *   :class:`thurible.progress.Tick`: Advance the progress bar
            and display any message passed.
        *   :class:`thurible.progress.NoTick`: Do not advance the
            progress bar but display the message passed as a
            temporary message. The temporary message will be replaced
            by the next message received.

        :param msg: A message sent by the application.
        :return: A :class:`str` object containing any updates needed to
            be made to the terminal display.
        :rtype: str
        """
        result = ''

        # If a tick is received, advance the progress bar.
        if isinstance(msg, Tick):
            if self._notick and self.max_messages:
                self.messages.popleft()
            self._notick = False
            self.progress += 1
            if self.max_messages:
                self._add_message(msg.message)
                self._wrapped_width = -1
            result += self._make_display()

        # If a notick is received, update the status messages but
        # don't advance the progress bar.
        elif isinstance(msg, NoTick) and self.max_messages:
            if self._notick:
                self.messages.popleft()
            self._notick = True
            self._add_message(msg.message)
            self._wrapped_width = -1
            result += self._make_display()

        return result

    # Private helper methods.
    def _add_message(self, msg) -> None:
        if self.timestamp:
            stamp = datetime.now() - self._t0
            mins = stamp.seconds // 60
            secs = int(stamp.seconds % 60)
            msg = f'{mins:0>2}:{secs:0>2} {msg}'
        self.messages.appendleft(msg)

    def _make_display(self) -> str:
        result = ''
        height = 1 + self.max_messages
        y = self.inner_y
        y += self._align_v('middle', height, self.inner_height)
        x = self.content_x
        result += self.term.move(y, x) + self.progress_bar
        y += 1

        if self.max_messages:
            result += self._visible_messages(x, y)

        return result

    def _visible_messages(self, x: int, y: int) -> str:
        result = ''
        width = self.content_width
        for i, line in zip(range(self.max_messages), self.lines):
            result += f'{self.term.move(y + i, x)}{line:<{width}}'
        return result



"""
dialog
~~~~~~

A dialog for terminal applications.
"""
from typing import Sequence, Optional

from blessed.keyboard import Keystroke

from thurible.menu import Option
from thurible.panel import Content, Title


# Common dialog options.
cont = (Option('Continue', ''),)
yes_no = (
    Option('Yes', 'y'),
    Option('No', 'n'),
)


# Classes.
class Dialog(Content, Title):
    """Create a new :class:`thurible.Dialog` object. This class displays
    a message to the user and offers pre-defined options for the
    user to chose from. As a subclass of :class:`thurible.panel.Content`
    and :class:`thurible.panel.Title`, it can also take those parameters
    and has those public methods and properties.

    :param message_text: The text of the prompt to be displayed to
        the user.
    :param options: The options the user can chose from. This is a
        sequence of :class:`thurible.Option` objects.
    :return: None.
    :rtype: NoneType.
    """
    def __init__(
        self,
        message_text: str,
        options: Sequence[Option] = yes_no,
        *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.message_text = message_text
        self.options = options

        # Defined action keys.
        self.register_key('KEY_ENTER', self._select)
        self.register_key('KEY_LEFT', self._select_left)
        self.register_key('KEY_RIGHT', self._select_right)
        for option in self.options:
            hotkey = f"'{option.hotkey}'"
            self.register_key(hotkey, self._hotkey)
        self._selected = len(self.options) - 1

    def __str__(self) -> str:
        result = super().__str__()

        result += self.message

        height = self.inner_height
        width = self.inner_width
        y = self._align_v('bottom', 1, height) + self.inner_y
        for i, option in enumerate(self.options[::-1]):
            opt_text = ''
            if i == len(self.options) - 1 - self._selected:
                opt_text += self.term.reverse
            name = option.name
            length = len(name) + 2
            x = self._align_h('right', length, width) + self.inner_x
            opt_text += f'{self.term.move(y, x)}[{name}]'
            if i == len(self.options) - 1 - self._selected:
                opt_text += self.term.normal
            result += opt_text
            width -= length + 1

        return result

    # Properties
    @property
    def message(self) -> str:
        """
        The message as a string that could be used to update the terminal.

        :return: A :class:`str` object.
        :rtype: str
        """
        wrapped = self.term.wrap(self.message_text, width=self.inner_width)
        length = len(wrapped)
        y = self._align_v('middle', length, self.inner_height) + self.inner_y
        x = self.inner_x
        result = ''
        for i, line in enumerate(wrapped):
            result += f'{self.term.move(y + i, x)}{line}'
        return result

    # Public methods.
    def action(self, key: Keystroke) -> tuple[str, str]:
        """Act on a keystroke typed by the user.

        :param key: A :class:`blessed.keyboard.Keystroke` object representing
            the key pressed by the user.
        :return: A :class:`tuple` object containing two :class:`str` objects.
            The first string is any data that needs to be sent to the
            application. The second string contains any updates needed
            to be made to the terminal display.
        :rtype: tuple
        """
        # These are the results that are returned.
        data = ''
        update = ''

        if repr(key) in self._active_keys:
            handler = self._active_keys[repr(key)]
            data = handler(key)
        else:
            data = str(key)

        if not data:
            height = self.inner_height
            width = self.inner_width
            y = self._align_v('bottom', 1, height)
            for i, option in enumerate(self.options[::-1]):
                opt_text = ''
                if i == len(self.options) - 1 - self._selected:
                    opt_text += self.term.reverse
                name = option.name
                length = len(name) + 2
                x = self._align_h('right', length, width)
                opt_text += f'{self.term.move(y, x)}[{name}]'
                if i == len(self.options) - 1 - self._selected:
                    opt_text += self.term.normal
                update += opt_text
                width -= length + 1

        return data, update

    # Private action handlers.
    def _hotkey(self, key: Optional[Keystroke] = None) -> str:
        """Select the option assigned to the hot key."""
        hotkeys = [option.hotkey for option in self.options]
        self._selected = hotkeys.index(str(key))
        return ''

    def _select(self, key: Optional[Keystroke] = None) -> str:
        """Return the name of the selected option."""
        return self.options[self._selected].name

    def _select_left(self, key: Optional[Keystroke] = None) -> str:
        """Select the next option to the left."""
        if self._selected > 0:
            self._selected -= 1
        return ''

    def _select_right(self, key: Optional[Keystroke] = None) -> str:
        """Select the next option to the right."""
        if self._selected < len(self.options) - 1:
            self._selected += 1
        return ''



"""
thurible
~~~~~~~~~~~

Managers for the data displays.
"""
from collections import deque
from dataclasses import dataclass
from queue import Queue
from typing import Optional

from blessed import Terminal

from thurible import messages as tm
from thurible.panel import Panel
from thurible.dialog import Dialog
from thurible.util import get_terminal


# Manager.
def queued_manager(
    q_to: Queue,
    q_from: Queue,
    term: Optional[Terminal] = None,
    displays: Optional[dict] = None
) -> None:
    """Manage a terminal display by sending and receiving
    :class:`thurible.messages.Message` objects through
    :class:`queue.Queue` objects.

    .. warning::
        :func:`thurible.queued_manager` is intended to be run in
        its own thread or process. If you try to run it synchronously
        with the rest of your application, the loop will prevent your
        application from completing execution. This is why it is a
        "queued" manager.

    :param q_to: A queue for messages the program sends to the manager.
    :param q_from: A queue for messages the manager sends to the program.
    :param term: An instance of `blessed.Terminal` used to interact with
        the terminal.
    :param displays: (Optional.) Storage for the panels the program may
        want the manager to display.
    :return: None.
    :rtype: NoneType

    Usage::

        >>> from queue import Queue
        >>> from threading import Thread
        >>> from thurible import get_terminal, queued_manager
        >>> from thurible.messages import End
        >>>
        >>> # Create a queue to send messages to the manager.
        >>> q_in = Queue()
        >>>
        >>> # Create a queue to receive messages from the manager.
        >>> q_out = Queue()
        >>>
        >>> # Get a terminal instance for the manager to use.
        >>> term = get_terminal()
        >>>
        >>> # Run the manager in a separate thread.
        >>> T = Thread(target=queued_manager, args=(q_in, q_out, term))
        >>> T.start()
        >>>
        >>> # End the thread running the queued_manager.
        >>> msg = End('Ending.')
        >>> q_in.put(msg)
    """
    # Set up.
    if term is None:
        term = get_terminal()
    if displays is None:
        displays = {}
    showing: str = ''
    history: deque[str] = deque(maxlen=100)
    farewell = ''
    reason = ''
    exception: Optional[Exception] = None
    last_height = term.height
    last_width = term.width

    # Program loop.
    with term.fullscreen(), term.cbreak(), term.hidden_cursor():
        while True:
            if showing and displays[showing].height == last_height:
                if term.height != last_height:
                    displays[showing].height = term.height
                    last_height = term.height
                    print(term.clear, end='', flush=True)
                    print(str(displays[showing]), end='', flush=True)
            if showing and displays[showing].width == last_width:
                if term.width != last_width:
                    displays[showing].width = term.width
                    last_width = term.width
                    print(term.clear, end='', flush=True)
                    print(str(displays[showing]), end='', flush=True)

            try:
                (
                    displays, showing, end, farewell, reason, history
                ) = check_messages(
                    q_to,
                    q_from,
                    displays,
                    showing,
                    history
                )
                if end:
                    break
                check_input(q_from, displays, showing, term)

            except Exception as ex:
                reason = 'Exception.'
                exception = ex
                break

    # After exiting full screen mode, print the farewell message and
    # inform the program the manager is ending..
    if farewell:
        print(farewell)
    q_from.put(tm.Ending(reason, exception))


# Manager core functions.
def check_input(
    q_from: Queue,
    displays: dict[str, Panel],
    showing: str,
    term: Terminal
) -> None:
    """Check if input from the user was received and act on any
    received.
    """
    key = term.inkey(timeout=.01)
    if key:
        update = ''
        data = str(key)
        if showing and isinstance(displays[showing], Panel):
            data, update = displays[showing].action(key)
        if update:
            print(update, end='', flush=True)
        if data:
            msg = tm.Data(data)
            q_from.put(msg)


def check_messages(
    q_to: Queue,
    q_from: Queue,
    displays: dict[str, Panel],
    showing: str,
    history: deque[str]
) -> tuple[dict[str, Panel], str, bool, str, str, deque[str]]:
    """Check if messages from the program were received and act on any
    received.
    """
    end = False
    reason = ''
    farewell = ''
    if not q_to.empty():
        msg = q_to.get()

        # End the manager.
        if isinstance(msg, tm.End):
            farewell = msg.text
            reason = 'Received End message.'
            end = True

        # Prove the manager is still responding.
        elif isinstance(msg, tm.Ping):
            pong = tm.Pong(msg.name)
            q_from.put(pong)

        # Display a stored panel.
        elif isinstance(msg, tm.Show):
            if showing:
                history.appendleft(showing)
            showing = msg.name
            print(str(displays[showing]), end='', flush=True)

        # Check what panel is currently displayed.
        elif isinstance(msg, tm.Showing):
            shown = tm.Shown(msg.name, showing)
            q_from.put(shown)

        # Store a panel for display.
        elif isinstance(msg, tm.Store):
            displays[msg.name] = msg.display

        # Remove a panel from storage.
        elif isinstance(msg, tm.Delete):
            del displays[msg.name]

        # Check what panels are currently stored.
        elif isinstance(msg, tm.Storing):
            stored = tm.Stored(
                msg.name,
                tuple(key for key in displays)
            )
            q_from.put(stored)

        # Show an alert.
        elif isinstance(msg, tm.Alert):
            height = len(get_terminal().wrap(
                msg.text,
                width=int(displays[showing].width * 0.6)
            ))
            rel_height = (height + 3) / displays[showing].height
            displays[msg.name] = Dialog(
                message_text=msg.text,
                options=msg.options,
                title_text=msg.title,
                frame_type='light',
                panel_align_h='center',
                panel_align_v='middle',
                panel_relative_height=rel_height,
                panel_relative_width=0.6,
                height=displays[showing].height,
                width=displays[showing].width
            )
            history.appendleft(showing)
            showing = msg.name
            print(str(displays[showing]), end='', flush=True)

        # Dismiss the alert.
        elif isinstance(msg, tm.Dismiss) and msg.name == showing:
            history.appendleft(showing)
            showing = history[1]
            print(str(displays[showing]), end='', flush=True)

        # Send unrecognized messages to the showing panel.
        else:
            update = displays[showing].update(msg)
            if update:
                print(update, end='', flush=True)

    return displays, showing, end, farewell, reason, history



"""
log
~~~

An object for displaying a history of updating messages, such as a
log.
"""
from collections import deque
from dataclasses import dataclass
from typing import Optional, Sequence

from thurible.panel import Content, Message, Title


# Available update message.
@dataclass
class Update(Message):
    """Create a new :class:`thurible.log.Update` object. This
    object is a command message used to instruct the currently
    displayed :class:`thurible.Log` to add the text given in the
    message.

    :param text: The message to add to the panel.
    :return: None.
    :rtype: NoneType
    """
    text: str


# Class.
class Log(Content, Title):
    """Create a new :class:`thurible.Log` object. This class displays
    messages from the application in "last in first out" (LIFO)
    format. It's intended for situations were you want to provide
    the user a rolling display of status messages. As a subclass of
    :class:`thurible.panel.Content` and :class:`thurible.panel.Title`,
    it can also take those parameters and has those public methods
    and properties.

    :param content: (Optional.) A sequence of strings to display
        in the panel when it is first displayed in the terminal.
        The first item in the sequence is considered the most
        recent.
    :param maxlen: (Optional.) The total number of entries the
        :class:thurible.Log will store. This is used to allow the
        terminal window to be resized without causing the loss of
        any messages. It's not intended for the user to be able to
        scroll to view messages that have rolled off the terminal.
    :return: None.
    :rtype: NoneType
    """
    def __init__(
        self,
        content: Optional[Sequence[str]] = None,
        maxlen: int = 50,
        *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.maxlen = maxlen
        if content is None:
            content = deque(maxlen=self.maxlen)
        elif not isinstance(content, deque):
            d: deque[str] = deque(maxlen=self.maxlen)
            for item in content:
                d.appendleft(item)
            content = d
        self.content = content

        self._wrapped_width = -1

    def __str__(self) -> str:
        """Return a string that will draw the entire panel."""
        # Set up.
        inner_height = self.inner_height
        y = self.inner_y
        x = self.inner_x
        result = super().__str__()

        # Write the contents of the log.
        result += self._visible(
            self.lines,
            self.inner_height,
            self.inner_x,
            self.inner_y
        )
        return result

    # Properties.
    @property
    def lines(self) -> list[str]:
        """The lines of text available to be displayed in the panel
        after they have been wrapped to fit the width of the
        interior of the panel. A message from the application may
        be split into multiple lines.

        :return: A :class:`list` object containing each line of
            text as a :class:`str`.
        :rtype: list
        """
        width = self.inner_width
        if width != self._wrapped_width:
            wrapped = []
            for line in self.content:
                wrapped.extend(self.term.wrap(line, width=width))
            self._lines = wrapped
            self._wrapped_width = width
        return self._lines

    # Public methods.
    def update(self, msg: Message) -> str:
        result = super().update(msg)

        if isinstance(msg, Update):
            self.content.appendleft(msg.text)
            self._wrapped_width = -1
            result += self.clear_contents()
            result += self._visible(
                self.lines,
                self.inner_height,
                self.inner_x,
                self.inner_y
            )

        return result

    # Private helper methods.
    def _visible(
        self,
        lines: Sequence[str],
        height: int,
        x: int,
        y: int
    ) -> str:
        result = ''
        for line in lines[:height]:
            result += self.term.move(y, x) + line
            y += 1
        return result



"""
__init__
~~~~~~~~

Initialization for the `thurible` package.
"""
import sys as _sys

# Subscripting of type is not supported before Python 3.9.
if _sys.version_info[:2] < (3, 9):
    msg = 'Thurible requires Python 3.9 or higher.'
    raise ImportError(msg)

from blessed import Terminal
from blessed.keyboard import Keystroke

from thurible.dialog import Dialog
from thurible.eventmanager import event_manager
from thurible.log import Log
from thurible.menu import Menu, Option
from thurible.progress import Progress
from thurible.splash import Splash
from thurible.table import Table
from thurible.text import Text
from thurible.textdialog import TextDialog
from thurible.thurible import queued_manager
from thurible.eventmanager import event_manager
from thurible.util import get_queues, get_terminal
